function [CLRate, ClassTable] = ClassifyData(A, ns, theta, samples)
% CLASSIFYDATA Classification by Maximum-A-Posteriori decision rule
%   [CLRate, ClassTable] = CLASSIFYDATA(A, ns, theta, samples) classifies
%   samples according to the Maximum-A-Posterior decision rule.
%
%   The input and output parameters are as follows:
%    * A: adjacency matrix of the Bayesian network to use
%    * ns: vector of sizes of the nodes in network, i.e. number of discrete 
%       values the nodes can take
%    * theta: probabilities of the network
%    * samples: matrix of samples (each column is a sample)
%
%    * CLRate: classification rate
%    * ClassTable: Classification Table
%
% --
% The provided software may be used free of charge for research purposes.
% For other uses and further support please contact Franz Pernkopf,
% pernkopf@tugraz.at.
%
% References:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek, 
%  "Maximum Margin Bayesian Network Classifiers," 
%  IEEE Transactions on Pattern Analysis and Machine Intelligence,
%  vol. 99, no. PrePrints, 2011
%
% Authors:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek
%
% Version:
%  1.0 August 24th, 2011
%

nrNodes = length(ns);

%% Create array with start and end indices of CPTs for certain variables
%% over all parent states
start_indices = zeros(nrNodes+1,1);
sidx = 1;
eidx = 0;
for k=1:nrNodes,
    if k > 1,
        sidx = eidx + 1;
    end

    parents = logical(A(:,k));
    nrParentStates = prod(ns(parents));

    eidx = eidx + ns(k) * nrParentStates;
    
    start_indices(k) = sidx;
end
start_indices(end) = eidx + 1;

%% Set up data structures
ClassTable = zeros(ns(1)); % Table with wrong classifications
nrSamples = size(samples,2); % number of samples to classify
ClassRate = 0; % Classification rate

%% Calculate joint probability of samples for possible classes
JointProb = ones(nrSamples, ns(1));

for j=1:nrNodes,
    for c = 1:ns(1),
        fam = [j, family(A,j)];
        
        values = samples(fam,:);
        [ind] = find(fam == 1);
        values(ind,:)=c;
        if ns(j) ~= 1
           sVect = theta(start_indices(j):(start_indices(j+1)-1));
           b = subv2ind(ns(fam)', values');
           JointProb(:,c) = JointProb(:,c).*(sVect(b));
        end
    end
end
        
%% Check which samples are classified correctly
for sidx = 1:nrSamples,
   [~,ind] = max(JointProb(sidx,:));
   
   if length(ind) > 1,
       error('Two classes with equal probabilty.\n');
   end
   
   if samples(1,sidx) == ind,
       ClassRate = ClassRate+1;
   end
   ClassTable(samples(1,sidx),ind) = ClassTable(samples(1,sidx),ind)+1;
end

%% Compute the classification rate
CLRate = ClassRate / nrSamples * 100;
