function [M, gradM]= OptimizationMM_TPAMI(beta, A, ns, IndexSet, IndexSet2, start_indices, samples,sset,kappa,lambda)
% OPTIMIZATIONMM_TPAMI Calculates the margin objective
%   [M, gradM] = OPTIMIZATIONCLL_TPAMI(beta, A, ns, IndexSet, ...
%      IndexSet2, start_indices, samples, sset, kappa, lambda) calculates 
%   the maximum-margin (MM) objectve and its gradient with respect to the 
%   supplied log-proabilities.
%
%   The input and output parameters are as follows:
%    * beta: vector of log-probabilities
%    * A: adjacency matrix of the Bayesian network to use
%    * ns: vector of sizes of the nodes in network, i.e. number of discrete 
%       values the nodes can take
%    * IndexSet, IndexSet2, start_indices, sset: Precalculated indices for
%       shorter runtimes of the code
%    * samples: matrix of samples (each column is a sample)
%    * kappa, lambda: tuning parameters of the algorithm (see publication)
%    
%    * negCLL: negative CLL of the training samples under the
%       log-probabilities beta
%    * gradCLL: gradient of the CLL with respect to the log-probabilities
%       beta
%
% --
% The provided software may be used free of charge for research purposes.
% For other uses and further support please contact Franz Pernkopf,
% pernkopf@tugraz.at.
%
% References:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek, 
%  "Maximum Margin Bayesian Network Classifiers," 
%  IEEE Transactions on Pattern Analysis and Machine Intelligence,
%  vol. 99, no. PrePrints, 2011
%
% Authors:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek
%
% Version:
%  1.0 August 24th, 2011
%

% Turn off Division-By-Zero Warnings
warning off MATLAB:divideByZero

%% Settings
eta = 5; % parameter of the softmax function

%% Define frequently used quantities
nrSamples = size(samples, 2);
nrNodes = length(ns);
nrClasses = ns(1);

%% renormalize probabilities
theta = exp(beta);
theta = normalize(A, ns, theta);

%% Compute M(Theta)

% (1) Compute sample probabilities using the current CPDs
JointProb = ones(nrSamples,nrClasses);
for j = 1:nrNodes,
    for c = 1:ns(1),
        JointProb(:,c) = JointProb(:,c) .* theta(IndexSet(:,j,c));
    end
end
LogJointProb = log(JointProb); % logarithm of joint probability

V = zeros(size(JointProb));

%% Compute the margin and the weight for each sample
idx = sub2ind(size(LogJointProb), 1:nrSamples, samples(1,:));
LM = LogJointProb(idx);

for m = 1:nrSamples,
    LJPsset = LogJointProb(m,sset(:,m));
    mLJPsset = max(LJPsset);
    LM(m) = LM(m) - mLJPsset;
    MV = ceil(mLJPsset);
    AA = LJPsset - MV;
    V(m,sset(:,m)) = (exp(AA).^eta)./(sum(exp(AA)).^eta);
end

if any(~isfinite(V)),
    warning('MMLearning:InfiniteElements', 'There are non-finite Elements');
    LM(isnan(LM)) = 0;
    LM(LM == -Inf) = -1000;
    V(isnan(V)) = 0;
end

%% Calculate objective
LMS = LM * lambda;

% Determine partitioning of the samples according to their margin
indS3 = find(LMS >= 1);
indS1 = find(LMS <= 1-2*kappa);
indS2 = setdiff(1:length(LM),union(indS3,indS1));

% calculate objective M(B|S) [Eq. (6)]
LMTotal = sum(LMS(indS1)) + kappa*length(indS1) ...
        + length(indS2) - 1/(4*kappa) * sum(((LMS(indS2)-1).^2)) ...
        + length(indS3);

M = -LMTotal; % objective (to be minimized)

% calculate weights [Eq. (8)]
weight = zeros(nrSamples,1);
weight(indS1) = lambda;
weight(indS2) = (-lambda/(2*kappa))*(LMS(indS2)-1);
weight(indS3) = 0;

% VW is the weighted V matrix
VW = V .* repmat(weight, 1, size(V,2));

%% Compute the gradient gradM(beta)
gradM = zeros(length(theta),1);
sidx = 1;
for j = 1:nrNodes,
    fam = [j, family(A,j)];
    sizes = ns(fam);
    nrParentStates = prod(sizes(2:end));
    
    indices = subv2ind(sizes', samples(fam,:)'); % each row of data' is a case

    count = zeros(prod(sizes),1);
    count(1:max(indices)) = accumarray(indices, weight);
    countsA = myreshape(count, sizes);
        
    if j == 1, % class node
        countsB = sum(VW)';
    else % non-class nodes
        countsB=zeros(size(count));
        for z = 1:nrClasses,
            indi = IndexSet2(:, j, z);  
            countsB(1:max(indi)) = countsB(1:max(indi))+accumarray(indi,VW(:,z));
        end
    end
    countsB = myreshape(countsB, sizes);
    
    Term1 = reshape(countsA-countsB, sizes(1), nrParentStates);
    
    Params = reshape(theta(start_indices(j):(start_indices(j+1)-1)), sizes(1), nrParentStates);

    Term2=(ones(size(Term1,1),1)*sum(Term1,1)).*(-1*Params);
    BG = Term1 + Term2;
    
    gradM(sidx:sidx + numel(BG) - 1) = reshape(-1*BG,numel(BG),1);

    sidx = sidx + numel(BG);
end



