function [theta] = learnMLParameters(A, ns, samples)
% LEARNMLPARAMETERS   Learns Maximum-Likelihood parameters 
%   LEARNMLPARAMETERS(A, ns, samples) determines the maximum likelihood
%   (ML) parameters of the Bayesian Network whose structure is given by the
%   adjacency matrix A. The sizes (number of discrete values) of the nodes
%   is specified by the entries of the vector ns. The matrix samples
%   contains the samples from which the parameters are estimated.
%
%   The input and output parameters are as follows:
%    * A: adjacency matrix of the Bayesian network to use
%    * ns: vector of sizes of the nodes in network, i.e. number of discrete 
%       values the nodes can take
%    * samples: matrix of samples (each column is a sample)
%
%    * theta: ML parameters
%
% --
% The provided software may be used free of charge for research purposes.
% For other uses and further support please contact Franz Pernkopf,
% pernkopf@tugraz.at.
%
% References:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek, 
%  "Maximum Margin Bayesian Network Classifiers," 
%  IEEE Transactions on Pattern Analysis and Machine Intelligence,
%  vol. 99, no. PrePrints, 2011
%
% Authors:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek
%
% Version:
%  1.0 August 24th, 2011
%

%% (1) calculate length of parameter vector
len = 0;
for n=1:length(ns),
    parents = logical(A(:,n));
    nrParentStates = prod(ns(parents));
    
    len = len + ns(n) * nrParentStates;
end

theta = zeros(len, 1);

sidx = 1;
for n=1:length(ns),
    parents = logical(A(:,n));
    nrParentStates = prod(ns(parents));

    nodes = [n, family(A,n)];

    eidx = (sidx+nrParentStates*ns(n)-1);
    theta(sidx:eidx) = hist(subv2ind(ns(nodes)', samples(nodes,:)'), 1:(nrParentStates*ns(n)));
    sidx = sidx + nrParentStates * ns(n);
end

%% Transform frequency counts to probabilities (if for certain CPDs there
%% have not been any observations, set the correspondiung CPD to a uniform
%% distribution)
sidx = 1;
for n=1:length(ns),
    parents = logical(A(:,n));
    nrParentStates = prod(ns(parents));
    
    for i=1:nrParentStates,
        eidx = sidx + ns(n) - 1;
        
        if all(theta(sidx:eidx) == 0), 
            theta(sidx:eidx) = 1; % set to uniform distribution
        end
        
        theta(sidx:eidx) = theta(sidx:eidx) / sum(theta(sidx:eidx));
        
        sidx = sidx + ns(n);
    end
end

end
