% Maximum Margin Bayesian Network Classifiers example script
%
% The provided software may be used free of charge for research purposes.
% For other uses and further support please contact Franz Pernkopf,
% pernkopf@tugraz.at.
%
% References:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek, 
%  "Maximum Margin Bayesian Network Classifiers," 
%  IEEE Transactions on Pattern Analysis and Machine Intelligence,
%  vol. 99, no. PrePrints, 2011
%
% Authors:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek
%
% Version:
%  1.0 August 24th, 2011
%

%% (*) clear environment
clear all
clc

%% Values to use to rediscover Table 1 and Table 2 from the publication
%% mentioned in the header

% MNIST
%               NB      TAN-CMI         TAN-OMI-CR      TAN-CR
% kappa         0.025   0.075           0.25            0.05
% lambda        0.2     0.1             0.075           0.1
% iter (MM)     47      39              33              34
% iter (CLL)    43      17              24              21

% USPS
%               NB      TAN-CMI         TAN-OMI-CR      TAN-CR
% kappa         0.075   0.05            0.01            0.075
% lambda        0.05    0.07            0.06            0.05
% iter (MM)     29      28              17              21
% iter (CLL)    33      53              30              43

%% (0) Specify simulation 

% dataset to use
dataset = 'USPS';
%dataset = 'MNIST';

% network structure to use
structure = 'NB';
%structure = 'TAN-CMI';
%structure = 'TAN-OMI-CR';
%structure = 'TAN-CR';

kappa = 0.075;
lambda = 0.05;
iter = 29;
iterCLL = 33;

% smoothing parameter
epsilon = 1e-5;

%% (1) load training and test data
% training data: USPS_Train.mat
% test data: USPS_Test.mat
% both contain matrices with samples that are organized as follows:
%   each column contains one training samples
%   each row is a specific attribute, where the first-row is the class
%     variable

% load training set
load(['datasets/', dataset, '_Train', '.mat']);
TrainData = Samples;

% load test set
load(['datasets/', dataset, '_Test', '.mat']);
TestData = Samples;

%% (2) Specify structure of graphical network by an adjacency matrix and
%% the number of values each attribute can take

struct_file = ['structures/', dataset, '_', structure, '.mat'];

% load adjecency matrix
load(struct_file);
nrNodes = size(A, 1);

% set up node sizes ns
ns = max(TrainData, [], 2);

%% dump problem information
nrCPTentries = 0;
for n = 1:nrNodes,
    nrParentStates = prod(ns(logical(A(:,n))));
    nrCPTentries = nrCPTentries + nrParentStates*ns(n);
end

fprintf(1, '>> Problem information\n');
fprintf(1, 'Training Data: %d samples\n', size(TrainData, 2));
fprintf(1, 'Test Data: %d samples\n', size(TestData, 2));
fprintf(1, 'Number of CPT-entries: %d entries\n\n', nrCPTentries);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ML learning %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% (3) Calculate ML parameters from training data
fprintf(1, '>> Learning ML parameters...\n');
tic;
thetaML = learnMLParameters(A, ns, TrainData);
% Smooth probabilities and renormalize
thetaML(thetaML == 0) = epsilon; 
thetaML = normalize(A, ns, thetaML);
tp = toc;
fprintf(1, '<< (took %e s)\n\n', tp);

%% (4) classify data using the obtained parameters
fprintf(1, '>> Classification using ML parameters...\n');
tic;
[CLRateTrain, CLTableTrain] = ClassifyData(A, ns, thetaML, TrainData);
[CLRateTest, CLTableTest] = ClassifyData(A, ns, thetaML, TestData);
fprintf(1, '* Training set: %3.2f %%\n', CLRateTrain);
fprintf(1, '* Test set: %3.2f %%\n', CLRateTest);
tp = toc;
fprintf(1, '<< (took %e s)\n\n', tp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% CLL learning %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% (3) Train the network
tic;
fprintf(1, '>> Learning CLL parameters...\n');
thetaCLL = learnCGCLLParameters(A, ns, thetaML, TrainData, iterCLL);
tp = toc;
fprintf(1, '<< (took %e s)\n\n', tp);

%% (4) classify data using the obtained parameters
fprintf(1, '>> Classification using CLL parameters...\n');
tic;
[CLRateTrain, CLTableTrain] = ClassifyData(A, ns, thetaCLL, TrainData);
[CLRateTest, CLTableTest] = ClassifyData(A, ns, thetaCLL, TestData);
fprintf(1, '* Training set: %3.2f %%\n', CLRateTrain);
fprintf(1, '* Test set: %3.2f %%\n', CLRateTest);
tp = toc;
fprintf(1, '<< (took %e s)\n\n', tp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% MM learning %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% (3) Train the network
tic;
fprintf(1, '>> Learning MM parameters...\n');
thetaMM = learnCGMMParameters(A, ns, thetaML, TrainData, iter, kappa, lambda);
tp = toc;
fprintf(1, '<< (took %e s)\n\n', tp);

%% (4) classify data using the obtained parameters
fprintf(1, '>> Classification using MM parameters...\n');
tic;
[CLRateTrain, CLTableTrain] = ClassifyData(A, ns, thetaMM, TrainData);
[CLRateTest, CLTableTest] = ClassifyData(A, ns, thetaMM, TestData);
fprintf(1, '* Training set: %3.2f %%\n', CLRateTrain);
fprintf(1, '* Test set: %3.2f %%\n', CLRateTest);
tp = toc;
fprintf(1, '<< (took %e s)\n\n', tp);
